/***** rank: rank.c *****/

/* Written by Seth Fogarty, Rice University, USA                          */
/* Based on code written by Laurent Doyen, ULB, Belgium                   */
/* Portions copyright (c) 2006  Laurent Doyen                             */
/* Portions copyright (c) 2011  Seth Fogarty                              */
/*                                                                        */
/* This program is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation; either version 2 of the License, or      */
/* (at your option) any later version.                                    */
/* This program is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */
/*                                                                        */
/* You should have received a copy of the GNU General Public License      */
/* along with this program; if not, write to the Free Software            */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA*/


#define Free free
#define Malloc malloc

#include "rank.h"

bool verbose=false;
bool generate_counterexample=false;
bool reallyVerbose=false;
bool csv=false;

const int NOT_LESS=0;
const int LESS_EQUAL=1;
const int LESS_STRICTLY=2;
bool containment=false;
bool owcty=false;


// Letters of the alphabet 
int num_letters=0;
int *sigma=NULL;


double r=2.0;  // Branhcing degree 
double f=0.5;  // Density of accepting states 
//The first automaton... which we will not complement 
int sizeA=1;
bool* initialA = NULL;
bool* acceptingA = NULL;
bool*** transitionsA = NULL; // Transitions over all letters 
bool report = false;

//the second, which we will 
int sizeB=0;
int maxRank=(-1);
bool* initialB = NULL;
bool* acceptingB = NULL;
bool*** transitionsB = NULL; // Transitions over all letters 

int largestRank=0;


List** postA;
List** postB;
List** preA;

int num_inserts = 0;
int num_removed = 0;
int num_subsumed = 0;

double start_time;

Counterexample* counterexample;


void set_params(double newr, double newf)
{
	r = newr;
	if (newf >= 0) f = newf;
	return;
}

void reset_time()
{
	start_time = (double) clock();
}

double get_elapsed_time()
{
	return (double) ((double) clock() - start_time) / CLOCKS_PER_SEC;
}

// SF: Constructors and destructors
Set emptySet() /*I hypothesize we only ever want sets of B. Do we sometimes want sets of A?*/ 
{
	int i;  
	Set s = (Set) malloc(sizeB*sizeof(int));
	for(i=0; i<sizeB; i++)
		s[i]=maxRank+1;
	return s;
}

// returns true if <s> is empty
bool isEmptySet(Set s)
{
	int i;
	for(i=0; i<sizeB; i++)
	{
		if (s[i] <= maxRank) return false;
	}
	return true;
}

int sizeSet(Set s)
{
	int i,size=0;
	for(i=0; i<sizeB; i++)
	{
		if (s[i] <= maxRank) size++;
	}
	return size;
}

void clearSet(Set s)
{
	int i;  
	for(i=0; i<sizeB; i++)
		s[i]=maxRank+1;
	return;
}

Set maximalSet()
{
	int i;  
	Set s = (Set) malloc(sizeB*sizeof(int));
	for(i=0; i<sizeB; i++)
		s[i]=0;
	return s;
}

Set initialSet()
{
	int v;
	Set s = (Set) malloc(sizeB*sizeof(int));
	for(v=0; v<sizeB; v++)
	{
		if (initialB[v]) s[v] = maxRank;
		else s[v] = maxRank+1;
	}
	return s;
}


Set copySet(Set s)
{
	int i;  
	Set result =(Set) malloc(sizeB*sizeof(int));
	for(i=0; i<sizeB; i++)
		result[i] = s[i];
	return result;
}

//Triple* createTriple(State a, Set s, Set o)
////{
	////Triple* result = (Triple*) malloc(sizeof(Triple));
	////result->a = a;
	//result->s = s;
	//result->o = o;
	//return result;
//}

Pair* createPair(Set s, Set o)
{
	Pair* result = (Pair*) malloc(sizeof(Pair));
	//result->a = -1;
	result->s = s;
	result->o = o;
	return result;
}


Pair* copyPair(Pair* t)
{
	Pair* result = (Pair*) malloc(sizeof(Pair));
	//result->a = t->a;
	result->s = copySet(t->s);
	result->o = copySet(t->o);
	return result;
}


Antichain* createAntichain()
{
	Antichain* result = (Antichain*) malloc(sizeof(Antichain));
	result->pair=NULL;
	result->next=NULL;
	return result;
}

AntichainVector createAntichainVector()
{
	Antichain* result;
	result = (Antichain*) malloc(sizeA*sizeof(Antichain));
	int i;
	for(i = 0; i < sizeA; i++)
	{
		result[i].pair = NULL;
		result[i].next = NULL; 
	}
	return result;
}

void deleteSet(Set s)
{
	if (s != NULL) {
		free(s);
		s = NULL;
	}
}

void deletePair(Pair* p)
{
	if (p != NULL)
	{
		deleteSet(p->s);
		deleteSet(p->o);
		free(p);
		p = NULL;
	}

}

void deleteAntichain(Antichain* Q)
{
	Antichain* Old;
	while(Q != NULL)
	{
		deletePair(Q->pair);
		Old = Q; 
		Q=Q->next;
		free(Old);
		Old = NULL;
	}
}

void deleteAntichainVector(AntichainVector QS)
{
	int i;
	for(i = 0; i < sizeA; i++)
	{
		assert(QS[i].pair == NULL);
		deleteAntichain(QS[i].next);
	}
	free(QS);
}

/*SF: This goes through and deletes all the antichain structs that were holding
 * the pairs together, but leaves the pairs intact. If you do not use R after
 * appendAntichain, call this. */

void deleteAntichainSpine(AntichainVector QS)
{
	int i;
	Antichain* Old, *Q;
	for(i=0;i<sizeA;i++)
	{
		Q=QS[i].next;
		while(Q != NULL)
		{
			Old = Q; 
			Q=Q->next;
			free(Old);
		}
	}
  free(QS);
}

//y=[[s,s],[s,owe]]  # y=top
/* SF: The top antichain is no longer small: Specifically, triples with
 * different a's are incomprable. Thus we now have a 2*sizeA top antichain. */

AntichainVector topAntichainVector()
{
	State i;
	AntichainVector y; 
	Antichain* cur;
	y = createAntichainVector();
	for(i = 0; i < sizeA; i++)
	{
		cur = &y[i];
		cur->next=createAntichain();
		cur->next->pair = createPair(maximalSet(), maximalSet());
		cur->next->next = createAntichain();
		cur->next->next->pair = createPair(maximalSet(), emptySet());
	}
	return y;
}



/*SF: Similarly, our initial antichain is now duplicated for every initial
 * state in A. It would be nice to know if the sets created really need to be
 * recreated each time: if they are write only (which I suspect they are), then
 * we can greatly simplify this and the above via memory-sharing.*/

AntichainVector initialAntichainVector()
{
	State i;
	Antichain* y;
	Antichain* cur;
	if(containment)
	{
		y = createAntichainVector();
		for(i = 0; i < sizeA; i++)
		{
			if(initialA[i])
			{
				cur = &y[i];
				cur->next = createAntichain();
				cur->next->pair= createPair(initialSet(), emptySet());
				cur = cur->next;
			}
		}
	}
	else
	{
		y = createAntichain();
		y->next = createAntichain();
		y->next->pair = createPair(initialSet(), emptySet());
	}
	return y;
}

//SF: End constructors and destructors

int min(int x, int y)
{
	if (x<y) return x;
	return y;
}


// returns the least odd number k such that n<=k
int ceilOdd(int n)
{
	if (n%2 == 1)
		return n;
	else
		return n+1;
}

// returns the least even number k such that n<=k
int ceilEven(int n)
{
	if (n%2 == 0)
		return n;
	else
		return n+1;
}


//SF: Show functions



void printWord(Word* W)
{
	if(W == NULL) {
		printf("\"\"");
	}
	else {
		printf("%d", W->letter);
		for(W=W->next; W != NULL; W=W->next)
		{
			printf(" %d",W->letter);
		}
	}
}


void printList(List list)
{
	int i;
	for(i=0; list[i] != -1; i++)
	{
		printf("%d ", list[i]);
	}
}

void printAllPost()
{
	int o, i;

	printf("\n***   Automaton A   ***\n");
	for(o=0; o<num_letters; o++)
	{
		printf("\n***\n***    sigma=%d    ***\n***\n",o);
		for(i=0; i<sizeA; i++)
		{
			printf("%d: ", i);
			printList(postA[o][i]);
			printf("\n");
		}
	}
	printf("\n***   Automaton B   ***\n");
	for(o=0; o<num_letters; o++)
	{
		printf("\n***\n***    sigma=%d    ***\n***\n",o);
		for(i=0; i<sizeB; i++)
		{
			printf("%d: ", i);
			printList(postB[o][i]);
			printf("\n");
		}
	}
}


void printSet(Set s)
{
	int i;
	if (s!= NULL)
	{
		printf("{");
		for(i=0; i<sizeB; i++)
		{
			if (i!=0) printf(", ");
			printf("%d: %d", i, s[i]);
		}
		printf("} ");
	}
	else
	{
		printf("Error ! Null pointer Set\n");
	}
}


void printSmartSet(Set s)
{
	int i;
	bool first = true;
	if (s != NULL)
	{
		for(i=0; (i<sizeB) && (s[i] > maxRank); i++)
			;
		printf("{");
		for(; i<sizeB; i++)
		{
			if (s[i] <= maxRank)
			{
				if (!first) printf(", ");
				printf("%d[%d]", i, s[i]);
				first=false;
			}
		}
		printf("}");
	}
	else
	{
		printf("Error ! Null pointer Set\n");
	}

}

void printPair(Pair* pair)
{
	if (pair != NULL)
	{
		//if (pair->a != -1) 
		//{
	//		printf("State in A: %d\n", pair->a);
	//	}
		printSet(pair->s);
		printf("\n");
		printSet(pair->o);
	}
	else
	{
		printf("Error ! Null pointer to Pair\n");
	}
	printf("\n");
}

void printSmartPair(Pair* pair)
{
	if (pair != NULL)
	{
		//if (pair->a != -1)
		//	printf("A: %d\n", pair->a);
		printSmartSet(pair->s);
		printf(",\t");
		printSmartSet(pair->o);
		printf("\n--\n");
	}
	else
	{
		printf("Error ! Null pointer to Pair\n");
	}
	printf("\n");
}


void printAntichain(Antichain* Q)
{
	if(Q!=NULL)
	{
		for(Q=Q->next; Q != NULL; Q=Q->next)
		{
			printPair(Q->pair);
			printf("\n");
		}
	}
	printf("\n");
}

void printSmartAntichain(Antichain* Q)
{
	int n=0;;
	if(Q==NULL)
	{
		printf("---\n");
		return;
	}
	for(Q=Q->next; Q != NULL; Q=Q->next)
	{
		if(verbose) printf("%d:  ", n++);
		printSmartPair(Q->pair);
		/*
		   if (!isEmptySet(Q->pair->o)) printf("  ");
		//for(i=0; i< 6*sizeB - 3*(sizeSet(Q->pair->s)+sizeSet(Q->pair->o)); i++) printf(" ");
		printf("{");
		for(v=0; v<sizeB; v++)
		{
		rank=0;
		if (Q->pair->o[v] <= maxRank)
		{
		rank = Q->pair->o[v];
		if (rank < Q->pair->s[v])
		{
		printf("Error in printSmartAntichain !\n Rank in o for %d: %d   in s: %d\n%d\n", v, rank, Q->pair->s[v], maxRank);
		exit(1);
		}
		}
		else if (Q->pair->s[v] <= maxRank)          
		rank = Q->pair->s[v];

		if (v!=0) printf(", ");
		printf("%d: %d", v, rank);
		}
		printf("}");  
		printf("\n");
		*/
	}
	printf("---\n");
}

void printAntichainVector(AntichainVector QS)
{
	int i;
	for(i = 0; i < sizeA; i++)
	{
		assert(QS[i].pair == NULL);
		printf("Antichain %d:\n",i);
		printAntichain(&(QS[i]));
		printf("-----\n");
	}
}

//SF: End show functions


//SF: Begin example generating, reading, and printing


// returns a random number in [0,1]
double getRandom()
{
	//printf("%f", rand()/(RAND_MAX+1.0));
	return rand()/(RAND_MAX+1.0);
}


void deletePost(List** arr, int size)
{
	int i, o;
	//printf("arr: %p\n", arr);
	if (arr != NULL)
	{
		for (o=0;o<num_letters;o++)
		{
			// printf("arr[%d]: %p\n", o, arr[o]);
			if (arr[o] != NULL)
			{
				for(i=0;i<size;i++)
				{
					//    printf("arr[%d][%d]: %p\n", o, i, arr[o][i]);
					if (arr[o][i] != NULL)
					{
						free(arr[o][i]);
						arr[o][i] = NULL;
					}
				}
				free(arr[o]);
				arr[o] = NULL;
			}
		}
		free(arr);
	}
	//printf("end arr: %p\n", arr);
}

void deleteTransitions(bool*** transitions, int size)
{
	int i, o;
	if (transitions != NULL) 
	{
		for (o = 0; o < num_letters; o++)
		{
			if (transitions[o] != NULL)
			{
				for(i=0; i<size; i++)
				{
					free(transitions[o][i]);
					transitions[o][i]=NULL;

				}
				free (transitions[o]);
				transitions[o]=NULL;
			}
		}
		free(transitions);
		transitions=NULL;
	}
}



void deleteExample()
{
	if (acceptingA != NULL) {
		free(acceptingA);
		acceptingA = NULL;
	}
	if (initialA != NULL) {
		free(initialA);
		initialA = NULL;
	}
	deletePost(postA, sizeA);
	postA=NULL;
	deletePost(preA, sizeA);
	preA=NULL;
	deleteTransitions(transitionsA, sizeA);
	transitionsA=NULL;
	if (acceptingB != NULL) {
		free(acceptingB);
		acceptingB = NULL;
	}
	if (initialB != NULL) {
		free(initialB);
		initialB= NULL;
	}
	deletePost(postB, sizeB);
	postB=NULL;
	deleteTransitions(transitionsB, sizeB);
	transitionsB=NULL;
}


//SF: For now we will simply set A to be the universal set.
void generateExample(int numberOfStates)
{
	int o, v,w;
	int alpha;    // number of accepting states
	double threshold;
	deleteExample();
	sizeB=numberOfStates;
	//maxRank=2*n;
	if (num_letters == 0) num_letters = 2;

	acceptingB = (bool*) malloc(sizeB*sizeof(bool));
	alpha=0;
	for(v=0; v<sizeB; v++)
	{
		acceptingB[v] = (getRandom() < f);
		if (acceptingB[v]) alpha++;
	}
	if (maxRank == -1) maxRank=2*(sizeB - alpha);
	// see "On Complementing Nondeterministic Buechi Automata", S. Gurumurthy, O. Kupferman, F. Somenzi, M. Y. Vardi

	threshold = (1.0*r)/sizeB;
	//printf("%f", threshold);

	transitionsB = (bool***) malloc(num_letters*sizeof(bool**));
	for(o=0; o < num_letters; o++)
	{
		transitionsB[o]= (bool**) malloc(sizeB*sizeof(bool*));
		for(v=0; v<sizeB; v++)
		{
			transitionsB[o][v]= (bool*) malloc(sizeB*sizeof(bool));
			for(w=0; w<sizeB; w++)
			{
				transitionsB[o][v][w] = (getRandom() < threshold);
			}
		}
	}

	initialB = (bool*) malloc(sizeB*sizeof(bool));
	for(v=1; v<sizeB; v++)
	{
		initialB[v] = false;
	}
	initialB[0] = true;  

	postB= (List**) malloc(num_letters*sizeof(List*));

	sizeA=1;
	acceptingA = (bool*) malloc(sizeof(bool));
	acceptingA[0] = true;
	transitionsA = (bool***) malloc(num_letters*sizeof(bool**));
	for(o=0; o < num_letters; o++)
	{
		transitionsA[o]= (bool**) malloc(sizeof(bool*));
		transitionsA[o][0]= (bool*) malloc(sizeof(bool));
		transitionsA[o][0][0] = true;
	}

	initialA = (bool*) malloc(sizeof(bool));
	initialA[0] = true;  

	postA= (List**) malloc(num_letters*sizeof(List*));
	preA= (List**) malloc(num_letters*sizeof(List*));
	for(v = 0; v < num_letters; v++)
	{
		postA[v]=NULL;
		postB[v]=NULL;
		preA[v]=NULL;
	}
}


/*Returns a string that needs to be freed */
int nextLine(FILE* stream, char** buffer)
{
	*buffer=NULL;
	int nbytes = 0;
	int bytes_read;
	bool needline = true;
	while(needline)
	{
		bytes_read = readLine(stream, buffer, &nbytes);
		if (bytes_read == -1) unix_error("Failed reading a line");
		if ((bytes_read == 0) && ((*buffer)[0] == '#' || (*buffer)[0] == '\n' || (*buffer)[0] == '\0'))
		{
			Free(*buffer);
			nbytes = 0;
		}
		else
		{
			needline = false;
		}
	}
	return (bytes_read)?0:1;
}



int state_list_to_vector(char* buffer, bool* vector)
{
	char* tmp;
	int result = 0;
	int counter = 0;
	int state = 0;
	tmp = strtok(buffer, " \t"); //get first number
	while(tmp != NULL) {
		counter++;
		result = sscanf(tmp, "%d", &state);
		//if (result != 1) unix_error("Found non-number %s in initial/final string %s\n", tmp, buffer);
		if (result != 1) {
			printf("ERROR");
			unix_error("[Found non-number %s in initial/final string %s]p", tmp, buffer);
		}
		vector[state] = true;
		tmp = strtok(NULL, " \t"); //get next number
	}
	return counter;
}


void readExample(FILE* file)
{
	int i, j;
	int v,w,letter;
	int result;
	int alpha=0;
	int maxRank_t;
	char* input= NULL;
	//FILE* file=fopen(filename, "r");
	//if (file == NULL) {printf("Error:  File <%s> not found !\n", filename); exit(1);}
	deleteExample();


	//reading format
	nextLine(file, &input);
	result = sscanf(input,"%d",&i); 
	if (result != 1) {
		printf("ERROR");
		unix_error("[Parsing error on line %s]", input);
	}
	if (i == 3) 
	{
		containment=true;    
	}
	else if (i == 2)
	{
		containment=false;
	}
	else {
		printf("Error");
		unix_error("[Format %d not supported.]", i);
	}
	Free(input);

	nextLine(file, &input);
	result = sscanf(input,"%d",&num_letters);  //reading number of letters
	if (result != 1) unix_error("Parsing error on line %s\n", input);
	Free(input);

	if(containment)
	{
		//READING A 
		nextLine(file, &input);
		result = sscanf(input,"%d",&sizeA);  //Reading n
		if (result != 1) unix_error("Parsing error on line %s\n", input);
		Free(input);

		//alocate space
		initialA = (bool*) malloc(sizeA*sizeof(bool));
		acceptingA = (bool*) malloc(sizeA*sizeof(bool));
		for(v=0; v<sizeA; v++)
		{
			initialA[v] = false;
			acceptingA[v] = false;
		}
		transitionsA = (bool***) malloc(num_letters*sizeof(bool**));
		for(letter=0; letter < num_letters; letter++)
		{
			transitionsA[letter]= (bool**) malloc(sizeA*sizeof(bool*));
			for(v=0; v<sizeA; v++)
			{
				transitionsA[letter][v]= (bool*) malloc(sizeA*sizeof(bool));
				for(w=0; w<sizeA; w++)
				{
					transitionsA[letter][v][w] = false;
				}
			}
		}

		nextLine(file, &input);
		state_list_to_vector(input, initialA); //reading initial states
		Free(input);

		//reading tranistions states
		nextLine(file, &input);
		while(input[0] != '-')
		{
			result = sscanf(input,"%d %d %d",&v, &letter, &w);
			if (result != 3) unix_error("Parsing error on line %s", input);
			Free(input);
			nextLine(file, &input);
			if (letter < num_letters) transitionsA[letter][v][w] = true;
			else printf("Error ! Unknown letter: %d.\n", letter);
		}
		Free(input);

		//reading accepting states
		nextLine(file, &input);
		state_list_to_vector(input, acceptingA);
		Free(input);
		postA = (List**) malloc(num_letters*sizeof(List*));
		preA = (List**) malloc(num_letters*sizeof(List*));
	}

	//READING B
	nextLine(file, &input);
	result = sscanf(input,"%d",&sizeB);  //Reading n
	if (result != 1) unix_error("Parsing error on line %s", input);
	Free(input);

	//alocate space
	initialB = (bool*) malloc(sizeB*sizeof(bool));
	acceptingB = (bool*) malloc(sizeB*sizeof(bool));
	for(v=0; v<sizeB; v++)
	{
		initialB[v] = false;
		acceptingB[v] = false;
	}
	transitionsB = (bool***) malloc(num_letters*sizeof(bool**));
	for(letter=0; letter < num_letters; letter++)
	{
		transitionsB[letter]= (bool**) malloc(sizeB*sizeof(bool*));
		for(v=0; v<sizeB; v++)
		{
			transitionsB[letter][v]= (bool*) malloc(sizeB*sizeof(bool));
			for(w=0; w<sizeB; w++)
			{
				transitionsB[letter][v][w] = false;
			}
		}
	}

	j = nextLine(file, &input);
	if (j==0) {
		unix_error("Returned 0 reading initial values: %s", input);
	}
	state_list_to_vector(input, initialB); //reading initial states
	Free(input);

	//reading tranistions states
	nextLine(file, &input);
	while(input[0] != '-')
	{
		result = sscanf(input,"%d %d %d",&v, &letter, &w);
		if (result != 3) unix_error("Parsing error on line %s\n", input);
		Free(input);
		nextLine(file, &input);
		if (letter < num_letters) transitionsB[letter][v][w] = true;
		else printf("Error ! Unknown letter: %d.\n", letter);
	}
	Free(input);

	//reading accepting states
	j = nextLine(file, &input);
	if (j!=0) alpha=state_list_to_vector(input, acceptingB);
	Free(input);

	//Seeing if the rank is provided for us.
	if(nextLine(file, &input))
	{
		result = sscanf(input, "%d", &maxRank_t);
		if ((maxRank == -1) && (result == 1))
		{
			maxRank = maxRank_t;
			if (verbose) printf("Rank set to %d\n", maxRank);
		}
    Free(input);
	}
  Free(input);
  //printf("input: %s",input);
	if (maxRank == -1) 
	{ 
		maxRank=2*(sizeB - alpha);
		if (verbose) printf("Rank set to %d\n", maxRank);
	}

	postB = (List**) malloc(num_letters*sizeof(List*));
	for(v = 0; v < num_letters; v++)
	{
		if(containment)
		{
			preA[v]=NULL;
			postA[v]=NULL;
		}
		postB[v]=NULL;
	}
	//fclose(file);

	//insert maxrank stuff here
}



void writeContainmentExample(char* file)
{
	int v,w;
	double fval = f;
	Letter letter;
	FILE* f=NULL;
	bool first = true;
	f=fopen(file, "w");
	fprintf(f, "#Arc List Containment\n3\n");
	fprintf(f, "#Size: %d, r: %.1f, f: %.1f\n\n", sizeB, r, fval);
	fprintf(f, "#Language size\n%d\n", num_letters);
	fprintf(f, "#First automaton\n");
	fprintf(f, "#Automaton size\n%d\n", sizeA);

	fprintf(f, "#Initial states:\n");

	for(v=0; v<sizeA; v++)
	{
		if (initialA[v])
		{
			if (first) 
			{
				fprintf(f, "%d", v);
				first = false;
			}
			else
			{
				fprintf(f, " %d", v);
			}
		}
	}
	fprintf(f, "\n");
	first = true;

	fprintf(f, "#Transitions\n");
	for(letter=0; letter<num_letters; letter++)
	{
		for(v=0; v<sizeA; v++)
		{
			for(w=0; w<sizeA; w++)
			{
				if (transitionsA[letter][v][w]) fprintf(f, "%d %d %d\n", v, letter, w);
			}
		}
	}

	fprintf(f, "-\n");

	fprintf(f, "#Accepting states:\n");
	for(v=0; v<sizeA; v++)
	{
		if (acceptingA[v])
		{
			if (first) 
			{
				fprintf(f, "%d", v);
				first = false;
			}
			else
			{
				fprintf(f, " %d", v);
			}
		}
	}

	fprintf(f, "\n#Second automaton\n");
	fprintf(f, "#Automaton size\n%d\n", sizeB);

	fprintf(f, "#Initial states:\n");

	first = true;
	for(v=0; v<sizeB; v++)
	{
		if (initialB[v])
		{
			if (first) 
			{
				fprintf(f, "%d", v);
				first = false;
			}
			else
			{
				fprintf(f, " %d", v);
			}
		}
	}
	fprintf(f, "\n");
	first = true;

	fprintf(f, "#Transitions\n");
	for(letter=0; letter<num_letters; letter++)
	{
		for(v=0; v<sizeB; v++)
		{
			for(w=0; w<sizeB; w++)
			{
				if (transitionsB[letter][v][w]) fprintf(f, "%d %d %d\n", v, letter, w);
			}
		}
	}

	fprintf(f, "-\n");

	fprintf(f, "#Accepting states:\n");
	for(v=0; v<sizeB; v++)
	{
		if (acceptingB[v])
		{
			if (first) 
			{
				fprintf(f, "%d", v);
				first = false;
			}
			else
			{
				fprintf(f, " %d", v);
			}
		}
	}

	fprintf(f, "\n"); //This should SO not be necessary
	fclose(f);
	if (verbose) printf("Example written in <%s>\n", file);
}

void writeUniversalityExample(char* file)
{
	int v,w;
	double fval = f;
	Letter letter;
	FILE* f=NULL;
	bool first = true;
	f=fopen(file, "w");
#ifndef OLDPRINT
	fprintf(f, "#Arc List Universality\n2\n");
	fprintf(f, "#Size: %d, r: %.1f, f: %.1f \n\n", sizeB, r, fval);
	fprintf(f, "#Language size\n%d\n", num_letters);
	fprintf(f, "#Automaton size\n%d\n", sizeB);

	fprintf(f, "#Initial states:\n");
	for(v=0; v<sizeB; v++)
	{
		if (initialB[v])
		{
			if (first) 
			{
				fprintf(f, "%d", v);
				first = false;
			}
			else
			{
				fprintf(f, " %d", v);
			}
		}
	}
	fprintf(f, "\n");
	first = true;

	fprintf(f, "#Transitions\n");
#else
	fprintf(f, "%d\n", sizeB);
	for(v=0; v<sizeB; v++)
		if (acceptingB[v])
			fprintf(f, "%d\n", v);
	fprintf(f, "-\n");
#endif
	for(letter=0; letter<num_letters; letter++)
	{
		for(v=0; v<sizeB; v++)
		{
			for(w=0; w<sizeB; w++)
			{
#ifdef OLDPRINT
				if (transitionsB[letter][v][w]) fprintf(f, "%d\n%d\n%d\n", v, letter, w);
#else
				if (transitionsB[letter][v][w]) fprintf(f, "%d %d %d\n", v, letter, w);
#endif
			}
		}
	}
#ifndef OLDPRINT

	fprintf(f, "-\n");

	fprintf(f, "#Accepting states:\n");
	for(v=0; v<sizeB; v++)
	{
		if (acceptingB[v])
		{
			if (first) 
			{
				fprintf(f, "%d", v);
				first = false;
			}
			else
			{
				fprintf(f, " %d", v);
			}
		}
	}
#endif

	fprintf(f, "\n");
	fclose(f);
	if (verbose) printf("Example written in <%s>\n", file);
}

void writeExample(char* file)
{
	if (containment)
		writeContainmentExample(file);
	else
		writeUniversalityExample(file);
}

void printExample()
{
	writeExample("graph.aut");
}

//SF: end example generating, reading, and printing


// SF: Begin initialization functions.
/* SF: The result is a list, terminated by -1. Each element of the list is
 * something s can transition to. 
 */
List Post(bool a, int s, Letter sigma)
{
	int i,j;
	int n = a ? sizeA : sizeB;
	bool ***transitions = a ? transitionsA : transitionsB;
	List result = (List) malloc((n+1)*sizeof(int));
	i=0;
	for (j=0; j<n; j++)
	{
		if (transitions[sigma][s][j])
		{
			result[i] = j;
			i++;
		}
	}
	result[i] = -1;
	return result;
}

/*This computes the predecessor ranking for A from state s using letter sigma*/
List PreA(int s, Letter sigma)
{
	int i,j;
	int n = sizeA;
	List result = (List) malloc((n+1)*sizeof(int));
	i=0;
	for(j=0; j<n; j++)
	{
		if (transitionsA[sigma][j][s])
		{
			result[i]=j;
			i++;
		}
	}
	result[i] = -1;
	return result;
}

void computeAllPost()
{
	int i;
	int sigma;
	for(sigma = 0; sigma < num_letters; sigma++)
	{
		if(containment)
		{
			postA[sigma] = (List*) malloc(sizeA*sizeof(List));
			for(i=0; i<sizeA; i++)
			{
				postA[sigma][i] = Post(true,i,sigma);
			}  
			preA[sigma] = (List*) malloc(sizeA*sizeof(List));
			for(i=0; i<sizeA; i++)
			{
				preA[sigma][i] = PreA(i,sigma);
			}  
		}
		postB[sigma] = (List*) malloc(sizeB*sizeof(List));
		for(i=0; i<sizeB; i++)
		{
			postB[sigma][i] = Post(false,i,sigma);
		}  
	}
}

void initialize()
{
	sizeB=4;
	//maxRank=2*n;
	srand((unsigned int)time((time_t *)NULL));
	if(verbose) printf("\n");
}

//SF: End initialization

// returns true if <s1> is a subset of <s2>
// SF: Do not conflate not_less with greater_than!
// SUbset is evidently 'not defined somewhere with a lower rank'
bool Subset(Set s1, Set s2)
{
	int i;
	for(i=0; i<sizeB; i++)
	{
		if ((s1[i] <= maxRank) 
				&& (s2[i] > s1[i])) 
			return false;
	}
	return true;
}

bool EQset(Set s1, Set s2)
{
	int i;
	for(i=0; i<sizeB; i++)
	{
		if ((s1[i] <= maxRank) 
				&& (s2[i] != s1[i])) 
			return false;
	}
	return true;
}

bool StSubset(Set s1, Set s2)
{
	int v;
	bool result=false;
	for(v=0; v<sizeB; v++)
	{
		if ((s1[v] <= maxRank) && (s2[v] > s1[v])) return false;
		if ((s2[v] <= maxRank) && (s2[v] < s1[v])) result=true;
	}
	return result;
}

int setRelation(Set s1, Set s2)
{
	int v;
	int result=LESS_EQUAL;
	for(v=0; v<sizeB; v++)
	{
		if ((s1[v] <= maxRank) && (s2[v] > s1[v])) return NOT_LESS;
		if ((s2[v] <= maxRank) && (s2[v] < s1[v])) result=LESS_STRICTLY;
	}
	return result;
}


/*SF: Computes the set that is the maximum of the two given sets*/

Set intersectSet(Set s1, Set s2)
{
	int i=0;
	Set result = malloc(sizeB*sizeof(List));
	for(i=0; i<sizeB; i++)
	{
		result[i]=(s1[i]>s2[i]?s1[i]:s2[i]); //max s1[i] s2[i]
	}
	return result;
}

/*Might return null! */
Pair* intersectPair(Pair *t1, Pair *t2)
{
	Set s;
	Set o;
	Pair* result = NULL;
	//if(t1->a == t2->a)
	{ 
		if(isEmptySet(t1->o) && isEmptySet(t2->o))
		{
			o = emptySet();
			s = intersectSet(t1->s, t2->s); //Why do we return the (A, 0, 0) pair? Is this possibly useful?
			result = createPair(s, o);
		}
		else
		{
			o = intersectSet(t1->o, t2->o); 
			if (!isEmptySet(o))
			{
				s = intersectSet(t1->s, t2->s);
				assert(!isEmptySet(s)); // If o is non-empty, s had better be non-empty.
				result = createPair(s, o);
			}
			else
			{
				free(o);
			}
		}
	}
	return result;
}



int getMaxRank(AntichainVector QS)
{
	int i,v;
	int result=0;
	Antichain* Q;
	for(i=0;i<sizeA;i++)
	{
		for(Q=QS[i].next; Q != NULL; Q=Q->next)
		{
			for(v=0; v<sizeB; v++)
			{
				int rank = Q->pair->o[v];
				if (rank <= maxRank)
				{
					if (rank > result) result=rank;
				}
			}
			for(v=0; v<sizeB; v++)
			{
				int rank = Q->pair->s[v];
				if (rank <= maxRank)
				{
					if (rank > result) result=rank;
				}
			}

		}
	}
	return result;
}

void updateLargestRank(AntichainVector QS)
{
	int rank=getMaxRank(QS);
	if (rank > largestRank) largestRank = rank;
}


int sizeAntichainVector(AntichainVector QS)
{
	int i;
	Antichain* Q;
	int result=0;
	for(i=0;i<sizeA;i++)
	{
		for(Q=QS[i].next; Q != NULL; Q=Q->next)
		{
			result++;
		}
	}
	return result;
}

bool isEmptyAntichainVector(AntichainVector QS)
{
	int i;
	for(i =0; i < sizeA; i++)
	{
		if(QS[i].next != NULL) 
			return false;
	}
	return true;
}


//OPTIMIZATION POINT: Check s first or o first?
bool LEQpair(Pair* p, Pair* q)
{
	if (//p->a == q->a
			 Subset(p->s,q->s) 
			&&  Subset(p->o,q->o) 
			&&  (isEmptySet(p->o)?isEmptySet(q->o):!isEmptySet(q->o))) 
		//     (isEmptySet(p->o)==isEmptySet(q->o)) 
		return true;
	return false;
}

bool EQpair(Pair* p, Pair* q)
{
	if (EQset(p->s,q->s) && EQset(p->o,q->o))
		return true;
	return false;
}

//SF: Not used
bool LTpair(Pair* p, Pair* q)
{
	int r=setRelation(p->s,q->s);
	if (r==LESS_STRICTLY)
	{
		if ((Subset(p->o,q->o) && ( (isEmptySet(p->o) && isEmptySet(q->o)) || ((!isEmptySet(p->o)) && (!isEmptySet(q->o))) )) )
			return true;
	}
	else if (((r==LESS_EQUAL) && (StSubset(p->o,q->o)) && ( (isEmptySet(p->o) 
						&& isEmptySet(q->o)) || ((!isEmptySet(p->o)) && (!isEmptySet(q->o))) ))
			)
		return true;
	return false;
}

// removes in (Q) all the elements that are < (pair)
// SF: This function is not used in the current implementation.
//SF: This function is NOT TESTED
//void removeStrictlyLess(Antichain* Q, Pair* pair) 
//{
//  Antichain* previous =Q;
//  for(Q=Q->next; Q != NULL;)
//  {
//    if (LTpair(Q->pair, pair))
//    {
//      // remove Q->pair from Q
//      // printf("Remove from Q\n");
//      previous->next=Q->next;
//      Q->next=NULL;
//      deleteAntichain(Q);
//      Q=previous->next;
//    }
//    else
//    {
//      Q=Q->next;
//      previous = previous->next;
//    }    
//  }
//}


// SF: Back to working code.
#ifdef NOSUBSUMPTION
//change this to remove subsumption
int insertAntichain(Antichain* Q, Pair* pair)
{
	num_inserts++;
	assert(Q->pair == NULL);
	//Antichain* save=Q;
	Antichain* previous =Q;
	// printf("* * * Entering insertAntichain * * *\n");
	// printPair(pair);
	// printf("\n--------------\n");
	// printSmartAntichain(Q);
	//printf("\n");

	if (pair == NULL) {printf("Error\n"); exit(1);}

	for(Q=Q->next; Q != NULL;)
	{
		// printPair(Q->pair);
		if (EQpair(pair, Q->pair))
		{
			deletePair(pair);
			num_subsumed++;
			return -1;
		}
		else
		{
			Q=Q->next;
			previous = previous->next;
		}
	}
	previous->next = createAntichain(); //SF: Q is null  here, so safe. Note we did not insert above, only deleted.
	previous->next->pair = pair;
	//printf("* * * Leaving insertAntichain * * *\n");
	return 0;
}

#endif

#ifndef NOSUBSUMPTION
int insertAntichain(Antichain* Q, Pair* pair)
{
	num_inserts++;
	assert(Q->pair == NULL);
	//Antichain* save=Q;
	Antichain* previous =Q;
	//printf("* * * Entering insertAntichain * * *\n");
	//printPair(pair);
	//printf("\n--------------\n");
	//printAntichain(Q);
	//printf("\n");

	//if (pair == NULL) {printf("Error\n"); exit(1);}

	for(Q=Q->next; Q != NULL;)
	{
		// printPair(Q->pair);
		if (LEQpair(pair, Q->pair))
		{
			deletePair(pair);
			num_subsumed++;
			return -1;
		}
		else if (LEQpair(Q->pair, pair))
		{
			// remove Q->pair from Q
			// printf("Remove from Q\n");
			num_removed++;
			previous->next=Q->next;
			Q->next=NULL;
			deleteAntichain(Q); 
			Q=previous->next;
		}
		else
		{
			Q=Q->next;
			previous = previous->next;
		}
	}
	previous->next = createAntichain(); //SF: Q is null  here, so safe. Note we did not insert above, only deleted.
	previous->next->pair = pair;
	//printf("* * * Leaving insertAntichain * * *\n");
	return 0;
}
#endif

bool memberAntichain(Antichain* Q, Pair* pair)
{
	assert(Q->pair == NULL);
	assert(pair != NULL);
	for(Q=Q->next; Q != NULL;)
	{
		if (EQpair(pair, Q->pair))
		{
			return true;
		}
		else
		{
			Q=Q->next;
		}
	}
	return false;
}

int insertAntichainVector(AntichainVector QS, State Astate, Pair* pair)
{
	Antichain* Q;
	if(containment)
	{
		Q=&(QS[Astate]);
		return insertAntichain(Q, pair);
	}
	else
	{
		return insertAntichain(QS,pair);
	}
}

Antichain* intersectAntichain(Antichain* Q, Antichain* P)
{
	Antichain* Pstart=P;
	Antichain* result = createAntichain();
	Pair* intersection;
	for(Q=Q->next; Q != NULL; Q=Q->next)
	{
		P=Pstart;
		for(P=P->next; P != NULL; P=P->next)
		{
			assert(P->pair != NULL);
			assert(Q->pair != NULL);
			intersection=intersectPair(P->pair, Q->pair);
			if (intersection != NULL) insertAntichain(result, intersection);
		}
	}
	return result;
}

AntichainVector intersectAntichainVector(AntichainVector QS, AntichainVector PS)
{
	int i;
	Antichain *Q, *P, *QiP;
	AntichainVector result = createAntichainVector();
	for(i=0;i<sizeA;i++)
	{
		Q=&(QS[i]);
		P=&(PS[i]);
		QiP=intersectAntichain(Q,P);
		result[i].next = QiP->next;
		assert(QiP->pair == NULL);
		free(QiP);
	}
	return result;
}


/* SF: fix by pushing R = R -> next down into each branch, and using a
 * temporary variable to hold pre-increment R here, and then delete as last
 * step. 
 */

/*SF: I believe this is DESTRUCTIVE on R, but is not totally so. 
 *In the CPRE's, that is certainly what we want to do. It appears that in the
 * least fixed point, however, we want R to be intact but modified. */

/*This copies pointers to the pairs of R, thus after calling this function call only deleteAntichainSpine(R) */

void appendAntichain(Antichain* Q, Antichain*  R)
{
	Antichain* previous= R;
	Antichain* Old;
	for(R=R->next; R!=NULL;)
	{
		if (insertAntichain(Q, (R->pair)) == -1)
		{
			//printf("#");
			previous->next = R->next;  
			Old = R;
			R=R->next;
			free(Old); 
			/* we do NOT want to delete the antichain here, we have already freed
			 * R->pair, we don't want to free R-> next, so we simply free R.*/
		}
		else
		{
			R=R->next; 
			previous = previous->next;
		}
	}
}

void appendAntichainVector(AntichainVector QS, AntichainVector  RS)
{
	int i;
	for(i =0; i < sizeA; i++)
	{
		appendAntichain(&QS[i],&RS[i]);
	}
}



// SF: Begin of heavy hitting logic functions. Here is where I miss OCaml. 

/* SF: The boolvalue acceptA tells us if we should restrict ourselves to
 * pairs containing an accepting state in A. it is the analogue to
 * PrePairGoalA */

AntichainVector PrePairBase(Pair* p, State a, bool acceptA)
{
	Set preo;
	Set pres;
	Set o=p->o;
	Set s=p->s;
	AntichainVector result;
	int j, v, w, rank, temp;
	Letter sigma;

	result = createAntichainVector();
	preo = emptySet();
	pres = emptySet();

	for(sigma=0; sigma<num_letters; sigma++)
	{
		clearSet(preo); 
		for(v=0; v<sizeB; v++)
		{
			// printf("v = %d", v);
			rank = 0;
			for(j=0; postB[sigma][v][j] != -1; j++)
			{
				w=postB[sigma][v][j];
				if (acceptingB[w])
				{
					if (o[w] > rank)
					{
						rank = o[w];
					}
				}
				else
				{
					temp = min(ceilOdd(s[w]), o[w]);
					if (temp > rank)
					{
						rank = temp;
					}
				}
			}
			//if(v==12) printf("found v=12: rank=%d, acceptingB[v]=%d\n", rank, acceptingB[v]);
			if (acceptingB[v])
				if (rank%2 == 1) rank++;
			//if(v==12) printf("found v=12: rank now=%d\n", rank);
			preo[v] = rank;
		}
		//if(report && (sigma==1)) {printSet(preo); printf("\n");}

		/* the above generates pre-o, this generates pre-s if pre-o is non-empty. */
		if (!isEmptySet(preo))
		{
			clearSet(pres);
			for(v=0; v<sizeB; v++)
			{
				rank = 0;
				for(j=0; postB[sigma][v][j] != -1; j++)
				{
					w=postB[sigma][v][j];
					if (s[w] > rank)
						rank = s[w];
				}
				if (acceptingB[v])
					if (rank%2 == 1) rank++;
				pres[v] = rank;
			}
			/* The predecessors of a state a,s,o is any state(a',s',o') such that a'->a and blah blah */
			for(j=0; preA[sigma][a][j] != -1; j++)
			{
				temp = preA[sigma][a][j]; 
				if(!acceptA || acceptingA[temp])
				{
					insertAntichainVector(result, temp, createPair(copySet(pres), copySet(preo)));
				}
			}
		}


		for(j=0; preA[sigma][a][j] != -1; j++)
		{
			temp = preA[sigma][a][j]; 
			if(!acceptA || acceptingA[temp])
			{
				insertAntichainVector(result, temp, createPair(copySet(preo), emptySet())); 
			}
		}
	}
	free(pres);
	free(preo); 
	/*SF: if we can say pairs are never altered, then we can avoid this by
	 * replacing clearSet with free() emptySet()! But I think they are freed, so
	 * we have to copy as above. */ 

	//printf("Leaving PrePair...\n");
	return result;
}

Antichain* PrePairUniversality(Pair* p)
{
	Set preo;
	Set o;
	Set pres;
	Set s;
	Antichain* result;
	int j;
	int v,w;
	int rank;
	int temp;
	Letter sigma;
	//printf("Entering PrePair...\n");
	//printPair(p);
	//printf("...\n");
	// print s
	// print o
	// print "are the arguments."

	result = createAntichain();
	o=p->o;
	s=p->s;

	for(sigma=0; sigma<num_letters; sigma++)
	{
		preo = emptySet();
		for(v=0; v<sizeB; v++)
		{
			// printf("v = %d", v);
			rank = 0;
			for(j=0; postB[sigma][v][j] != -1; j++)
			{
				w=postB[sigma][v][j];
				if (acceptingB[w])
				{
					if (o[w] > rank)
					{
						rank = o[w];
					}
				}
				else
				{
					temp = min(ceilOdd(s[w]), o[w]);
					if (temp > rank)
					{
						rank = temp;
					}
				}
			}
			if (acceptingB[v])
				if (rank%2 == 1) rank++;
			preo[v] = rank;
		}

		if (!isEmptySet(preo))
		{
			pres = emptySet();
			for(v=0; v<sizeB; v++)
			{
				rank = 0;
				for(j=0; postB[sigma][v][j] != -1; j++)
				{
					w=postB[sigma][v][j];
					if (s[w] > rank)
						rank = s[w];
				}
				if (acceptingB[v])
					if (rank%2 == 1) rank++;
				pres[v] = rank;
			}

			insertAntichain(result, createPair(pres, copySet(preo)));
		}
		// if (checkRelation(o,s))
		// printf("This test was useful");
		// pause();
		insertAntichain(result, createPair(preo, emptySet())); 
	}

	//printf("Leaving PrePair...\n");
	return result;
}

AntichainVector PrePair(Pair* p, State a)
{
  if(containment)
  {
	  return PrePairBase(p, a, false);
  }
  else
  {
    return PrePairUniversality(p);
  }
}

//This does not affect Q in any way, use normal deleteAntichain
AntichainVector CPreSet(AntichainVector QS)
{
	AntichainVector result = createAntichainVector();
	AntichainVector cpreQ;
	Antichain* Q;
	int i;
	for(i = 0; i < sizeA; i++)
	{
		for(Q=QS[i].next; Q != NULL; Q=Q->next)
		{
			// if (DEBUG) {printf("CPre of "); printPair(Q->pair);}
			cpreQ = PrePair(Q->pair, i);
			appendAntichainVector(result, cpreQ);
			//SF: Need to free more than just cpreQ.
			// free(cpreQ);
			deleteAntichainSpine(cpreQ);
		}
	}
	return result;
}



/*SF: This works great for when we are intersecting with B's accepting states.
 * Guess what it doesn't work for? Yeah, A's. */

AntichainVector PrePairGoalB(Pair* p, State a)
{
	Set preo;
	Set o;
	Set s;
	AntichainVector result;
	int j;
	int v,w;
	int rank;
	int temp;
	Letter sigma;

	result = createAntichainVector();
	o=p->o;
	s=p->s;
	//a=p->a;
	preo = emptySet();

	for(sigma=0; sigma<num_letters; sigma++)
	{
		clearSet(preo);
		for(v=0; v<sizeB; v++)
		{
			// printf("v = %d", v);
			rank = 0;
			for(j=0; postB[sigma][v][j] != -1; j++)
			{
				w=postB[sigma][v][j];
				if (acceptingB[w])
				{
					if (o[w] > rank)
					{
						rank = o[w];
					}
				}
				else
				{
					temp = min(ceilOdd(s[w]), o[w]);
					if (temp > rank)
					{
						rank = temp;
					}
				}
			}
			if (acceptingB[v])
				if (rank%2 == 1) rank++;
			preo[v] = rank;
		}
		// if (checkRelation(o,s))
		// printf("This test was useful");
		// pause();
		if (containment)
		{
			for(j=0; preA[sigma][a][j] != -1; j++)
			{
				insertAntichainVector(result, preA[sigma][a][j], createPair(copySet(preo), emptySet())); 
			}
		}
		else
		{
			insertAntichainVector(result, 0, createPair(copySet(preo), emptySet())); 
		}
	}
	free(preo);
	return result;
}

Antichain* PrePairGoalA(Pair* p, State a)
{
	return PrePairBase(p, a, true);
}


//This does not affect Q in any way, use normal deleteAntichain
AntichainVector CPreGoalA(AntichainVector QS)
{
	AntichainVector result = createAntichainVector();
	AntichainVector cpreQ;
	Antichain* Q;
	int i;
	for(i = 0; i < sizeA; i++)
	{
		for(Q=QS[i].next; Q != NULL; Q=Q->next)
		{
			// if (DEBUG) {printf("CPre of "); printPair(Q->pair);}
			cpreQ = PrePairGoalA(Q->pair, i);
			appendAntichainVector(result, cpreQ);
			//SF: Need to free more than just cpreQ
			//free(cpreQ);
			deleteAntichainSpine(cpreQ);
		}
	}
	return result;
}

//This does not affect Q in any way, use normal deleteAntichain
AntichainVector CPreGoalB(AntichainVector QS)
{
	AntichainVector result = createAntichainVector();
	AntichainVector cpreQ;
	Antichain* Q;
	int i;
	for(i = 0; i < sizeA; i++)
	{
		for(Q=QS[i].next; Q != NULL; Q=Q->next)
		{
			// if (DEBUG) {printf("CPre of "); printPair(Q->pair);}
			cpreQ = PrePairGoalB(Q->pair, i);
			appendAntichainVector(result, cpreQ);
			//SF: Need to free more than just cpreQ
			//free(cpreQ);
			deleteAntichainSpine(cpreQ);
		}
	}
	return result;
}

bool LEQantichain(AntichainVector QS, AntichainVector RS)
{
	bool ok;
	int i;
	Antichain* Q;
	Antichain* R;

	for(i = 0; i < sizeA; i++)
	{
		for(Q=QS[i].next; Q != NULL; Q=Q->next)
		{
			ok=false;
			for(R=RS[i].next; R != NULL && ok==false; R=R->next)
			{
				if (LEQpair(Q->pair, R->pair))
				{
					ok = true;
				}
			}
			if (!ok)
			{
				return false;
			}
		}
	}
	return true;
}

bool DisjointAntichains(AntichainVector QS, AntichainVector RS)
{
	Antichain* R;
	Antichain* Q;
	Antichain* oldQ;
	int i;

	//we want to ensure that nothing in Q is covered by R.
	//Thus we look for a counterexample.

	for(i=0; i < sizeA; i++)
	{
		Q=&QS[i];
		oldQ=Q;
		for(Q=Q->next; Q != NULL; Q=Q->next)
		{
			for(R=RS[i].next; R != NULL; R=R->next)
			{
				if (LEQpair(Q->pair, R->pair))
				{
					return false;
				}
			}
			//once we are here we know that Q->pair is disjoint, and will forever be so.
			//Since we are decreasing the size of our set at eveyr iteration,
			//we may as well delete it.
			oldQ->next=Q->next;
			Q->next=NULL;
			deleteAntichain(Q);
			Q = oldQ; //and since we delete Q->pair, we don't need to increment oldQ.
		}
	}
	//if we never find anything in Q that is covered by 
	//something in R, we are good
	return true;
}

// returns LEQantichain({pair}, Q)

bool LEQpairAntichainVector(Pair* pair, AntichainVector QS)
{
	int i;
	Antichain* Q;
	for(i=0;i<sizeA;i++)
	{
		for(Q=QS[i].next; Q != NULL; Q=Q->next)
		{
			if (LEQpair(pair, Q->pair))
			{
				return true;
			}
		}
	}
	return false;
}


//This is TOTALLY destructive of the goal. After you call this function, goal
//is completely freed.
AntichainVector computeLeastFixedPoint(AntichainVector goal)
{
	AntichainVector x;
	AntichainVector old;
	AntichainVector frontier;

	frontier = goal;
	x=createAntichainVector();
	appendAntichainVector(x, frontier);

	if (reallyVerbose) printf("Computing Least Fixedpoint, entering\n");
	while(!isEmptyAntichainVector(frontier))
	{
		if (TIME_LIMIT > 0)
		{
			if (get_elapsed_time() > TIME_LIMIT) return x;
		}
		if (reallyVerbose) printf(".");
		old = frontier;
		frontier = CPreSet(frontier);  // memory leak ?? 
		if (reallyVerbose) printf("Computing Least Fixedpoint, adding\n");
		/*SF: Do we need to delete the old fronteir? I believe so, but how do we
		 * know which elements to delete? I think it is safe to delete everything...*/
		deleteAntichainSpine(old);
		if (reallyVerbose) printf("Appending.\n");

		appendAntichainVector(x, frontier);  // this modifies frontier 
		/*SF: Specifically, it removes things in frontier that already in x. so
		 * everything that is in frontier now will NOT be in it next pass. Nice. So
		 * our CPre will only get 'new' things. But when we insert something new,
		 * it stays on the frontier*/
	}
	deleteAntichainSpine(frontier);
	if (reallyVerbose) printf("Computing Least Fixedpoint, returning\n");
	return x;
}

//cannot be destructive on b
AntichainVector computeSymbolicReachability(AntichainVector b)
{
	AntichainVector y; 
	AntichainVector yOld;
	AntichainVector cprey; 
	AntichainVector fixed;

	y=b;
	yOld = NULL;
	do
	{
		if((yOld != NULL) && (yOld != b))
		{
			deleteAntichainVector(yOld);
		}
		yOld=y;
		cprey = CPreSet(y); 
		fixed = intersectAntichainVector(y,cprey);
		y=fixed;

	} while(!LEQantichain(yOld, y));
	if((yOld != NULL) && (yOld != b))
	{
		deleteAntichainVector(yOld);
	}
	return y;
}

void assertPostComputed()
{
	int v;
	assert(postB != NULL);
	if(containment)
	{
		assert(postA != NULL);
		assert(preA != NULL);
	}
	for(v = 0; v < num_letters; v++)
	{
		assert(postB[v]!=NULL);
		if(containment)
		{
			assert(postA[v]!=NULL);
			assert(preA[v]!=NULL);
		}
	}
	return;
}


//int testAntichainAndSet(Antichain* Q, AntichainVector QS)
//{
//	bool result = true;
//	int i;
//	Antichain* P;
//	result = result && (isEmptyAntichainVector(QS)==isEmptyAntichain(Q));
//	result = result && (sizeAntichainVector(QS)==sizeAntichain(Q));
//	result = result && (getMaxRankSet(QS)==getMaxRank(Q));
//	for(i=0; i<sizeA; i++) 
//	{
//		for(P=QS[i].next; P != NULL; P=P->next)
//			result = result && memberAntichain(Q, P->pair);
//	}
//
//	if(!result)
//	{
//		printf("Failure in comparing antichain and antichain set!\n");
//		printSmartAntichain(Q);
//		printf("--------------------------\n\n");
//		printAntichainVector(QS);
//	}
//	assert(result);
//	return 0;
//
//
//}

/*SF: This is computing the greatest fixed point. */
/*We return true if the containment HOLDS. Which is to say the complementary
 * automata is EMPTY. Thus we wait for the point that we can no longer reach
 * initial states. */

bool checkContainment(AntichainVector init, AntichainVector fixedPoint)
{
	int iteration = 0;
	assertPostComputed();

	if (reallyVerbose) printf("Init: \n");
	if (reallyVerbose) printAntichainVector(init);
	AntichainVector y; 
	AntichainVector yOld; 
	AntichainVector cpreyA; 
	AntichainVector cpreyB; 
	AntichainVector fixedA; 
	AntichainVector fixedB; 
	AntichainVector reachable;
	bool completed = false;
	cpreyA=NULL;

	y=topAntichainVector();
	yOld = NULL;
	if (reallyVerbose) printf("Iteration: %d, |y|= %d, maxRank=%d\n", iteration, sizeAntichainVector(y), getMaxRank(y) );
	if (reallyVerbose) printf("y: \n");
	if (reallyVerbose) printAntichainVector(y);
	do
	{
#ifdef OUTPUT
		printf(".\n");
#endif
		/*SF: Is there a memory leak here? Do we not need to free yold?*/
		if(yOld != NULL) deleteAntichainVector(yOld);
		yOld=y;
		if(owcty)
		{
			reachable = computeSymbolicReachability(y);
			y=reachable;
		}
		if(containment)
		{
			cpreyA = CPreGoalA(y); //SF: Pre(y) intersect accepting of A
		}
		cpreyB = CPreGoalB(y); //SF: Pre(y) intersect accepting of B
		if (reallyVerbose) printf("CPre Goal of B:\n");
		if (reallyVerbose) printAntichainVector(cpreyB);

		if (reallyVerbose) printf("Entering least fixed point.\n");
		if(containment)
		{
			fixedA = computeLeastFixedPoint(cpreyA);
			fixedB = computeLeastFixedPoint(cpreyB);
			y=intersectAntichainVector(fixedA, fixedB);  //Do I now need to free fixedA and fixedB?
			deleteAntichainVector(fixedA);
			deleteAntichainVector(fixedB);
		}
		else
		{
			y=computeLeastFixedPoint(cpreyB); //SF: mu x . Pre(x) union cprey
		}


		iteration++;
		if (reallyVerbose) printf("Iteration: %d, |y|= %d, maxRank=%d\n", iteration, sizeAntichainVector(y), getMaxRank(y) );
		if (reallyVerbose) printf("y-old: \n");
		if (reallyVerbose) printAntichainVector(yOld);
		if (reallyVerbose) printf("y: \n");
		if (reallyVerbose) printAntichainVector(y);

#ifdef OUTPUT
		printf("\n");
		printf("\n");
#endif

		if (DisjointAntichains(init, y)) 
		{
			//printf("QUIT EARLY");
			completed = true;
		}

		if (TIME_LIMIT > 0)
		{
			if (get_elapsed_time() > TIME_LIMIT)
			{
				completed = true;
			}
		}

	} while((!LEQantichain(yOld, y)) && !completed);
	deleteAntichainVector(yOld);
	fixedPoint->next=y->next; 
	completed |= DisjointAntichains(init, y); 
	/*if disjoint, then there is no lasso, and the automata
	  is empty, and we ARE contained */
	if (verbose) printf("Value upon return: %d\n", completed);

	if (!completed && generate_counterexample)
	{
		counterexample= generateCounterexample(init, y);
	}
	free(y);
	return completed;
}


void buchiSimple()
{
	AntichainVector fixedPoint = createAntichainVector();
	AntichainVector init = initialAntichainVector();
	bool isContained = checkContainment(init, fixedPoint);
	deleteAntichainVector(init);
	deleteAntichainVector(fixedPoint);

	if(containment)
	{
		if (isContained) printf("Contained");
		else printf("Not Contained");
	}
	else
	{
		//if (isContained) printf("Contained");
		//else printf("Not Contained");
		if (isContained) printf("Universal");
		else printf("Not Universal");
	}
	if(!isContained && generate_counterexample) 
	{
		printf("\nCounterexample: ");
		printWord(counterexample->prefix);
		printf(" (");
		printWord(counterexample->loop);
		printf(")^omega\n");
	}
	if(csv) printf(", ");
	else printf("\n");

	return;
}


void setmaxRank(int a)
{
	if (verbose) printf("Setting maximum rank to %d\n", a);
	maxRank = a;
}

int getmaxRank()
{
	return maxRank;
}

void print_help()
{
	printf("\nrank-based Buechi automata containment and universality checker, version 1.2.\n");
#ifdef NOSUBSUMPTION
	printf("Compiled without subsumption.\n");
#endif
	printf("Copyrights (c) 2006  Laurent Doyen and 2011  Seth Fogarty.\n");

	printf("Usage: rank [option..] file \n\n");


	printf("Valid options are:\n");
	printf("        -h                     Print this message and exit.\n");
	printf("        -c                     Generate and print a counterexample word.\n");
	printf("        -owcty                 Use OWCTY modification to Emerson-Lei universality algorithm.\n");
	printf("        -r [n]                 Use maximum rank n.\n");
	printf("        -v                     Be verbose in output.\n");
	printf("        -vv                    Be very verbose in output.\n");
	printf("        -csv                   Use commas instead of newlines.\n");
	printf("        -o [s] [r] [f] [n]     Generate (but do not analyze) n automata of size s,transition density r,\n");
	printf("                                   and acceptance density f. (Default s=50, r=2.0, f=0.5, n=1)\n");

	return;

}


//int unittest()
//{
//	FILE* file=fopen("test.aut", "r");
//	if (file == NULL) {printf("Error:  File <%s> not found !\n", "test.aut"); exit(1);}
//	readExample(file);
//	AntichainVector QS = NULL;
//	AntichainVector PS = NULL;
//	AntichainVector QPIS = NULL;
//	Antichain* Q = NULL;
//	Antichain* P = NULL;
//	Antichain* QPI = NULL;
//
//	printf(".");
//	QS = createAntichainVector();
//	Q = createAntichain();
//	testAntichainAndSet(Q,QS);
//	assert(isEmptyAntichainVector(QS));
//	assert(sizeAntichainVector(QS)==0);
//	deleteAntichainVector(QS);
//	deleteAntichain(Q);
//
//	printf(".");
//	QS = topAntichainVector();
//	Q = topAntichain();
//	testAntichainAndSet(Q,QS);
//	assert(!isEmptyAntichainVector(QS));
//	assert(sizeAntichainVector(QS)==10);
//
//	printf(".");
//	PS = initialAntichainVector();
//	P = initialAntichain();
//	testAntichainAndSet(P,PS);
//	assert(!isEmptyAntichainVector(PS));
//	assert(sizeAntichainVector(PS)==1);
//
//	printf(".");
//	QPIS = intersectAntichainVector(QS,PS);
//	QPI = intersectAntichain(Q,P);
//	testAntichainAndSet(QPI,QPIS);
//
//	printf(".");
//	appendAntichainVector(QS,QPIS);
//	appendAntichain(Q,QPI);
//	testAntichainAndSet(Q,QS);
//	deleteAntichainVector(QS);
//	deleteAntichain(Q);
//	deleteAntichainVectorSpine(QPIS);
//	deleteAntichainSpine(QPI);
//
//	printf(".");
//	Q=createAntichain();
//	QS=createAntichainVector();
//	appendAntichainVector(QS,PS);
//	appendAntichain(Q,P);
//	testAntichainAndSet(Q,QS);
//	deleteAntichainVector(QS);
//	deleteAntichain(Q);
//	deleteAntichainVectorSpine(PS);
//	deleteAntichainSpine(P);
//
//
//	printf(".\n");
//	return 0;
//}

int main(int argc, char **argv)
{
	int size=SIZE;
	int num=1;
	int i;
	char filename[100];
	double r = -1.0;
	double f = -1.0;
	int curloc, oldloc;
	reset_time();
	initialize();
	curloc=1;
	oldloc=0;
	while(curloc < argc && oldloc != curloc)
	{
		oldloc=curloc;
		if(strcmp(argv[curloc], "--mem")==0)
		{
			printf("%d bytes in a pointer\n",sizeof(void*));
			return 0;
		}

		if (strcmp(argv[curloc], "-o")==0)
		{
			if (argc >= 3) size = atoi(argv[2]);
			if (argc >= 4)
			{
				r = atof(argv[3]);
				if (argc >= 5) 
				{
					f = atof(argv[4]);
					if (argc == 6) num = atof(argv[5]);
				}
				set_params(r, f);
			}
			for(i = 0; i < num; i++)
			{
				generateExample(size);

				sprintf(filename, "graph%d.aut", i); 

				writeExample(filename);
			}
			return 0;
		}

		if ((strcmp(argv[curloc],"-h")==0) || (strcmp(argv[curloc],"--help")==0) || (strcmp(argv[curloc],"-help")==0))
		{
			print_help (); 
			return 0;
		}

		if (strcmp(argv[curloc], "-v")==0)
		{
			verbose = true; 
			curloc++;
		}
		else if (strcmp(argv[curloc], "-vv")==0)
		{
			reallyVerbose = verbose = true; 
			curloc++;
		}
		else if (strcmp(argv[curloc], "-c")==0)
		{
			generate_counterexample=true;
			curloc++;
		}
		else if (strcmp(argv[curloc], "-csv")==0)
		{
			csv = true;
			curloc++;
		}
		else if (strcmp(argv[curloc], "-owcty")==0)
		{
			owcty = true;
			curloc++;
		}
		else if (strcmp(argv[curloc], "-r")==0)
		{
			curloc++;
			setmaxRank(atoi(argv[curloc]));
			curloc++;
		}

	}
	FILE* file;
	if (argc > (curloc))
	{
        file=fopen(argv[curloc], "r");
		if (file == NULL) {printf("Error:  File <%s> not found !\n", argv[curloc]); exit(1);}
	}
	else
	{
		file = stdin;
	}

	readExample(file);
	computeAllPost();
	buchiSimple();
	if(verbose) printf("Duration of the computation: %f\n", get_elapsed_time());
#ifdef NOSUBSUMPTION
	if(verbose) printf("Inserts: %d, Inserts avoided from equality: %d(%.0f%%),\
			Removals during insertion: %d(%.0f%%)\n",
			num_inserts,num_subsumed,
			((float)num_subsumed*100)/(float)num_inserts,
			num_removed,
			((float)num_removed*100)/(float)num_inserts);
#else
	if(verbose) printf("Inserts: %d, Subsumed inserts: %d(%.0f%%), Removals from\
			subsumption: %d(%.0f%%)\n",
			num_inserts,num_subsumed,
			((float)num_subsumed*100)/(float)num_inserts,
			num_removed,
			((float)num_removed*100)/(float)num_inserts);
#endif 
	deleteExample();
#ifdef MEMORY
	mallocCount();
#endif
	//exit(1);

	return 0;
}
